from __future__ import annotations

from pathlib import Path
import runpy
import sys
from types import SimpleNamespace

import click
from click.testing import CliRunner
import pytest

from kanbus import cli
from kanbus.config_loader import ConfigurationError
from kanbus.hooks import HookEvent, HookExecutionError, HookPhase
from kanbus.issue_comment import IssueCommentError
from kanbus.issue_lookup import IssueLookupError
from kanbus.migration import MigrationError

from test_helpers import build_issue, build_project_configuration


def _run(args: list[str]) -> object:
    return CliRunner().invoke(cli.cli, args)


def test_run_lifecycle_hooks_helper_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    context = click.Context(click.Command("kanbus"))
    context.obj = None
    cli._run_lifecycle_hooks_for_context(
        context,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_LIST,
        root=tmp_path,
    )

    context.obj = {"beads_mode": False, "no_hooks": False, "no_guidance": False}
    monkeypatch.setattr(
        cli,
        "run_lifecycle_hooks",
        lambda **_k: (_ for _ in ()).throw(HookExecutionError("hook fail")),
    )
    with pytest.raises(click.ClickException, match="hook fail"):
        cli._run_lifecycle_hooks_for_context(
            context,
            phase=HookPhase.BEFORE,
            event=HookEvent.ISSUE_LIST,
            root=tmp_path,
        )


def test_dep_parser_and_beads_remove_error_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    context = click.Context(click.Command("dep"))
    context.obj = {}

    with context.scope(cleanup=False):
        with pytest.raises(click.ClickException, match="usage"):
            cli.dep.callback(())
        with pytest.raises(click.ClickException, match="tree requires an identifier"):
            cli.dep.callback(("tree",))
        with pytest.raises(click.ClickException, match="tree requires an identifier"):
            cli.dep.callback(("tree", ""))
        with pytest.raises(click.ClickException, match="depth must be a number"):
            cli.dep.callback(("tree", "kanbus-1", "--depth", "x"))

    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "build_dependency_tree", lambda *_a, **_k: {"ok": True})
    monkeypatch.setattr(cli, "render_dependency_tree", lambda *_a, **_k: "rendered")
    with context.scope(cleanup=False):
        cli.dep.callback(("tree", "kanbus-1", "--unknown", "value"))

    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _r: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        "kanbus.beads_write.remove_beads_dependency",
        lambda *_a, **_k: (_ for _ in ()).throw(
            cli.BeadsWriteError("beads remove fail")
        ),
    )
    with context.scope(cleanup=False):
        with pytest.raises(click.ClickException, match="beads remove fail"):
            cli.dep.callback(("kanbus-1", "remove", "blocked-by", "kanbus-2"))


def test_snyk_and_jira_configuration_error_callbacks(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _r: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(ConfigurationError("cfg fail")),
    )

    snyk_context = click.Context(cli.snyk_pull)
    with snyk_context.scope(cleanup=False):
        with pytest.raises(click.ClickException, match="cfg fail"):
            cli.snyk_pull.callback(False, None, None, None)

    jira_context = click.Context(cli.jira_pull)
    jira_context.obj = {"beads_mode": False, "no_hooks": False, "no_guidance": False}
    with jira_context.scope(cleanup=False):
        with pytest.raises(click.ClickException, match="cfg fail"):
            cli.jira_pull.callback(False)


def test_show_update_and_comment_remaining_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_resolve_beads_mode", lambda _ctx, _beads: (False, False))
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    config_calls = {"count": 0}

    def _load_config_show_then_ok(_p):
        config_calls["count"] += 1
        if config_calls["count"] == 1:
            raise ConfigurationError("cfg read fail")
        return build_project_configuration()

    monkeypatch.setattr(cli, "load_project_configuration", _load_config_show_then_ok)
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: SimpleNamespace(issue=build_issue("kanbus-1")),
    )
    monkeypatch.setattr(cli, "format_issue_for_display", lambda *_a, **_k: "shown")
    result_show = _run(["show", "kanbus-1"])
    assert result_show.exit_code == 0
    assert "shown" in result_show.output

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text, warnings=[], suggestions=[]),
    )
    monkeypatch.setattr(cli, "validate_code_blocks", lambda _t: None)
    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueLookupError("missing before")),
    )
    monkeypatch.setattr(cli, "update_issue", lambda **_k: build_issue("kanbus-1"))
    emitted: list[str] = []
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: emitted.append("emit"))
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    result_update_regular = _run(
        ["update", "kanbus-1", "--description", "desc", "--no-validate"]
    )
    assert result_update_regular.exit_code == 0
    assert emitted

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr("kanbus.beads_write.add_beads_comment", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "load_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(MigrationError("after missing")),
    )
    result_comment_beads_after_missing = _run(
        ["comment", "kanbus-1", "hello", "--no-validate"]
    )
    assert result_comment_beads_after_missing.exit_code == 0

    monkeypatch.setattr(
        "kanbus.beads_write.add_beads_comment",
        lambda *_a, **_k: (_ for _ in ()).throw(
            cli.BeadsWriteError("beads comment fail")
        ),
    )
    result_comment_beads_fail = _run(["comment", "kanbus-1", "hello", "--no-validate"])
    assert result_comment_beads_fail.exit_code != 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(
        cli,
        "add_comment",
        lambda **_k: (_ for _ in ()).throw(IssueCommentError("comment fail")),
    )
    result_comment_regular_fail = _run(
        ["comment", "kanbus-1", "hello", "--no-validate"]
    )
    assert result_comment_regular_fail.exit_code != 0
    assert "comment fail" in result_comment_regular_fail.output

    result_comment_required = _run(["comment", "kanbus-1"])
    assert result_comment_required.exit_code != 0
    assert "Comment text required" in result_comment_required.output


def test_create_update_move_delete_list_and_dep_remaining_paths(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_resolve_beads_mode", lambda _ctx, _beads: (False, False))
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "apply_text_quality_signals",
        lambda text: SimpleNamespace(text=text, warnings=[], suggestions=[]),
    )
    monkeypatch.setattr(cli, "validate_code_blocks", lambda _t: None)
    monkeypatch.setattr(cli, "emit_signals", lambda *_a, **_k: None)
    monkeypatch.setattr(cli, "format_issue_for_display", lambda *_a, **_k: "formatted")
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )

    issue = build_issue("kanbus-1")
    monkeypatch.setattr(
        cli,
        "create_issue",
        lambda **_k: SimpleNamespace(
            issue=issue, configuration=build_project_configuration()
        ),
    )
    result_create_regular = _run(
        ["create", "x", "--description", "desc", "--no-validate"]
    )
    assert result_create_regular.exit_code == 0

    monkeypatch.setattr(cli, "_resolve_beads_mode", lambda _ctx, _beads: (True, True))
    monkeypatch.setattr(cli, "create_beads_issue", lambda **_k: issue)
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    result_create_beads = _run(
        ["create", "x", "--description", "desc", "--no-validate"]
    )
    assert result_create_beads.exit_code == 0
    monkeypatch.setattr(cli, "_resolve_beads_mode", lambda _ctx, _beads: (False, False))

    show_config_calls = {"count": 0}

    def _show_config(_p):
        show_config_calls["count"] += 1
        if show_config_calls["count"] == 1:
            return build_project_configuration(beads_compatibility=True)
        return build_project_configuration()

    monkeypatch.setattr(cli, "load_project_configuration", _show_config)
    monkeypatch.setattr(cli, "load_beads_issue", lambda *_a, **_k: issue)
    result_show_beads = _run(["show", "kanbus-1"])
    assert result_show_beads.exit_code == 0

    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _r: tmp_path)
    monkeypatch.setattr(cli, "load_beads_issue", lambda *_a, **_k: issue)
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda *_a, **_k: (_ for _ in ()).throw(cli.ProjectMarkerError("no project")),
    )
    monkeypatch.setattr(cli, "update_beads_issue", lambda *_a, **_k: None)
    result_update_beads = _run(["update", "kanbus-1", "--description", "desc"])
    assert result_update_beads.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(ConfigurationError("cfg fail")),
    )
    monkeypatch.setattr(cli, "update_issue", lambda **_k: issue)
    result_move_regular = _run(["move", "kanbus-1", "task"])
    assert result_move_regular.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    result_move_beads_guard = _run(["move", "kanbus-1", "task"])
    assert result_move_beads_guard.exit_code != 0
    assert "not supported in beads mode" in result_move_beads_guard.output

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(beads_compatibility=True),
    )
    monkeypatch.setattr(
        cli,
        "load_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(MigrationError("missing")),
    )
    monkeypatch.setattr(cli, "delete_beads_issue", lambda *_a, **_k: None)
    result_delete = _run(["delete", "kanbus-1", "--yes"])
    assert result_delete.exit_code == 0

    monkeypatch.setattr(cli, "list_issues", lambda *_a, **_k: [issue])
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(cli, "format_issue_line", lambda *_a, **_k: "line")
    result_list_pm = _run(["list"])
    assert result_list_pm.exit_code == 0

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(ConfigurationError("list cfg fail")),
    )
    result_list_cfg = _run(["list"])
    assert result_list_cfg.exit_code != 0
    assert "list cfg fail" in result_list_cfg.output

    # Direct dep callback for successful depth parse branch.
    context = click.Context(click.Command("dep"))
    context.obj = {}
    monkeypatch.setattr(cli, "build_dependency_tree", lambda *_a, **_k: {"ok": True})
    monkeypatch.setattr(cli, "render_dependency_tree", lambda *_a, **_k: "rendered")
    with context.scope(cleanup=False):
        cli.dep.callback(("tree", "kanbus-1", "--depth", "2"))

    monkeypatch.setattr(
        cli,
        "get_configuration_path",
        lambda _r: (_ for _ in ()).throw(
            cli.ProjectMarkerError("project not initialized")
        ),
    )
    jira_context = click.Context(cli.jira_pull)
    jira_context.obj = {"beads_mode": False, "no_hooks": False, "no_guidance": False}
    with jira_context.scope(cleanup=False):
        with pytest.raises(click.ClickException, match='Run \\"kanbus init\\"'):
            cli.jira_pull.callback(False)


def test_cli_main_guard_line_executes() -> None:
    previous_argv = sys.argv[:]
    try:
        sys.argv = ["kanbus", "--help"]
        with pytest.raises(SystemExit):
            runpy.run_module("kanbus.cli", run_name="__main__")
    finally:
        sys.argv = previous_argv



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import beads_write

from test_helpers import build_issue


def _write_jsonl(path: Path, records: list[dict]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text("\n".join(json.dumps(r) for r in records) + "\n", encoding="utf-8")


def _record(issue_id: str, **overrides):
    base = {
        "id": issue_id,
        "title": issue_id,
        "description": "",
        "status": "open",
        "priority": 2,
        "issue_type": "task",
        "owner": "dev",
        "created_at": "2026-03-09T00:00:00Z",
        "created_by": "dev",
        "updated_at": "2026-03-09T00:00:00Z",
        "comments": [],
        "dependencies": [],
    }
    base.update(overrides)
    return base


def test_slug_sequence_and_slug_generation(monkeypatch: pytest.MonkeyPatch) -> None:
    beads_write.set_test_beads_slug_sequence(["abc", "xyz"])
    assert beads_write._next_beads_slug() == "abc"
    assert beads_write._next_beads_slug() == "xyz"
    assert beads_write._next_beads_slug() is None

    beads_write.set_test_beads_slug_sequence(["qqq"])
    assert beads_write._generate_slug() == "qqq"

    monkeypatch.setattr(beads_write.secrets, "choice", lambda alphabet: alphabet[0])
    beads_write.set_test_beads_slug_sequence(None)
    assert beads_write._generate_slug() == "aaa"


def test_basic_helpers() -> None:
    assert beads_write._beads_comment_uuid(
        "kanbus-1", "1"
    ) == beads_write._beads_comment_uuid("kanbus-1", "1")

    assert beads_write._derive_prefix({"kb-aaa", "kb-bbb"}) == "kb"
    with pytest.raises(beads_write.BeadsWriteError, match="invalid beads id"):
        beads_write._derive_prefix({"nohyphen"})

    assert beads_write._next_child_suffix({"kb-1.1", "kb-1.9", "other"}, "kb-1") == 10

    assert beads_write._generate_identifier({"kb-aaa"}, "kb", parent="kb-1") == "kb-1.1"


def test_generate_identifier_collision_exhaustion(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    existing = {"kb-aaa"}
    monkeypatch.setattr(beads_write, "_generate_slug", lambda: "aaa")
    with pytest.raises(
        beads_write.BeadsWriteError, match="unable to generate unique id"
    ):
        beads_write._generate_identifier(existing, "kb", parent=None)


def test_create_beads_issue_error_paths(tmp_path: Path) -> None:
    with pytest.raises(beads_write.BeadsWriteError, match="no .beads directory"):
        beads_write.create_beads_issue(tmp_path, "T", None, None, None, None, None)

    beads = tmp_path / ".beads"
    beads.mkdir()
    with pytest.raises(beads_write.BeadsWriteError, match="no issues.jsonl"):
        beads_write.create_beads_issue(tmp_path, "T", None, None, None, None, None)

    issues_path = beads / "issues.jsonl"
    issues_path.write_text("", encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="no beads issues available"):
        beads_write.create_beads_issue(tmp_path, "T", None, None, None, None, None)


def test_create_beads_issue_success_without_project_dir(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa")])

    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(beads_write, "_generate_slug", lambda: "bbb")
    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )

    issue = beads_write.create_beads_issue(
        tmp_path,
        "New",
        issue_type="task",
        priority=1,
        assignee="me",
        parent=None,
        description="d",
    )
    assert issue.identifier == "kb-bbb"
    records = beads_write._load_beads_records(issues_path)
    assert any(record.get("id") == "kb-bbb" for record in records)


def test_create_beads_issue_parent_not_found_and_event_rollback(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa")])

    monkeypatch.setattr(beads_write, "_generate_slug", lambda: "bbb")
    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")

    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.create_beads_issue(tmp_path, "x", None, None, None, "missing", None)

    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(
        beads_write,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )

    original = issues_path.read_text(encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="event fail"):
        beads_write.create_beads_issue(tmp_path, "x", None, None, None, None, None)
    assert issues_path.read_text(encoding="utf-8") == original


def test_create_beads_issue_parent_and_publish_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa"), _record("kb-aaa.1")])

    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(
        beads_write, "issue_created_payload", lambda _issue: {"ok": True}
    )
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "write_events_batch", lambda *_a: None)
    monkeypatch.setattr(
        beads_write, "load_beads_issue", lambda _r, _i: build_issue("kb-aaa.2")
    )
    published: list[tuple[str | None, str]] = []
    monkeypatch.setattr(
        beads_write,
        "publish_issue_mutation",
        lambda _r, _p, issue, event_id, topic: published.append(
            (event_id, issue.identifier)
        ),
    )

    issue = beads_write.create_beads_issue(
        tmp_path,
        "Child",
        issue_type=None,
        priority=None,
        assignee=None,
        parent="kb-aaa",
        description=None,
    )
    assert issue.identifier == "kb-aaa.2"
    records = beads_write._load_beads_records(issues_path)
    created = [r for r in records if r.get("id") == "kb-aaa.2"][0]
    assert created.get("dependencies")
    assert len(published) == 3


def test_create_beads_issue_load_beads_issue_failures_after_event_write(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa")])

    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(beads_write, "_generate_slug", lambda: "bbb")
    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(
        beads_write, "issue_created_payload", lambda _issue: {"ok": True}
    )
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "write_events_batch", lambda *_a: None)
    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("boom")),
    )

    published: list[str] = []
    monkeypatch.setattr(
        beads_write,
        "publish_issue_mutation",
        lambda _r, _p, issue, _event_id, _topic: published.append(issue.identifier),
    )

    issue = beads_write.create_beads_issue(
        tmp_path,
        "New",
        issue_type=None,
        priority=None,
        assignee=None,
        parent=None,
        description=None,
    )
    assert issue.identifier == "kb-bbb"
    # Falls back to publishing the initially built issue only.
    assert published == ["kb-bbb"]


def test_update_beads_issue_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa", labels=["a"])])

    monkeypatch.setattr(
        beads_write, "load_beads_issue", lambda _root, _id: build_issue("kb-aaa")
    )

    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.update_beads_issue(tmp_path, "missing")

    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )
    updated = beads_write.update_beads_issue(
        tmp_path,
        "kb-aaa",
        status="in_progress",
        title="New",
        description="D",
        priority=1,
        assignee="me",
        add_labels=["b"],
        remove_labels=["a"],
    )
    assert updated.identifier == "kb-aaa"

    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(
        beads_write,
        "build_update_events",
        lambda *_a: [SimpleNamespace(event_id="evt-1")],
    )
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(
        beads_write,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )

    original = issues_path.read_text(encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="event fail"):
        beads_write.update_beads_issue(tmp_path, "kb-aaa", set_labels=["x"])
    assert issues_path.read_text(encoding="utf-8") == original


def test_update_beads_issue_additional_error_and_publish_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    with pytest.raises(beads_write.BeadsWriteError, match="no .beads directory"):
        beads_write.update_beads_issue(tmp_path / "x", "a")

    beads = tmp_path / ".beads"
    beads.mkdir()
    with pytest.raises(beads_write.BeadsWriteError, match="no issues.jsonl"):
        beads_write.update_beads_issue(tmp_path, "a")

    issues_path = beads / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa")])

    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda _root, _id: (_ for _ in ()).throw(RuntimeError("lookup fail")),
    )
    with pytest.raises(beads_write.BeadsWriteError, match="lookup fail"):
        beads_write.update_beads_issue(tmp_path, "kb-aaa")

    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda _root, _id: (_ for _ in ()).throw(RuntimeError("after fail")),
    )
    with pytest.raises(beads_write.BeadsWriteError, match="after fail"):
        beads_write.update_beads_issue(tmp_path, "kb-aaa", title="t")

    lookup_calls = iter([build_issue("kb-aaa"), RuntimeError("after fail second")])

    def _load_beads_issue_second(_root: Path, _id: str):
        value = next(lookup_calls)
        if isinstance(value, Exception):
            raise value
        return value

    monkeypatch.setattr(beads_write, "load_beads_issue", _load_beads_issue_second)
    with pytest.raises(beads_write.BeadsWriteError, match="after fail second"):
        beads_write.update_beads_issue(tmp_path, "kb-aaa", title="t2")

    monkeypatch.setattr(
        beads_write, "load_beads_issue", lambda _root, _id: build_issue("kb-aaa")
    )
    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(beads_write, "build_update_events", lambda *_a: [])
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "write_events_batch", lambda *_a: None)
    published: list[str | None] = []
    monkeypatch.setattr(
        beads_write,
        "publish_issue_mutation",
        lambda _r, _p, _i, event_id, _t: published.append(event_id),
    )
    beads_write.update_beads_issue(tmp_path, "kb-aaa", title="new")
    assert published == [None]


def test_add_beads_comment_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("kb-aaa")])

    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.add_beads_comment(tmp_path, "missing", "a", "t")

    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )
    with pytest.raises(beads_write.BeadsWriteError, match="no project"):
        beads_write.add_beads_comment(tmp_path, "kb-aaa", "a", "t")

    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "load_beads_issue", lambda _root, _id: build_issue("kb-aaa")
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(beads_write, "comment_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(
        beads_write,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )

    original = issues_path.read_text(encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="event fail"):
        beads_write.add_beads_comment(tmp_path, "kb-aaa", "a", "t")
    assert issues_path.read_text(encoding="utf-8") == original


def test_add_beads_comment_missing_dirs_and_load_issue_failure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    with pytest.raises(beads_write.BeadsWriteError, match="no .beads directory"):
        beads_write.add_beads_comment(tmp_path / "x", "a", "u", "t")

    beads = tmp_path / ".beads"
    beads.mkdir()
    with pytest.raises(beads_write.BeadsWriteError, match="no issues.jsonl"):
        beads_write.add_beads_comment(tmp_path, "a", "u", "t")

    issues_path = beads / "issues.jsonl"
    _write_jsonl(issues_path, [_record("a")])
    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda _root, _id: (_ for _ in ()).throw(RuntimeError("load fail")),
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(beads_write, "comment_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "write_events_batch", lambda *_a: None)
    beads_write.add_beads_comment(tmp_path, "a", "u", "t")


def test_add_beads_comment_missing_comments_key_and_second_project_lookup_fails(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(
        issues_path,
        [
            {
                "id": "a",
                "title": "a",
                "description": "",
                "status": "open",
                "priority": 2,
                "issue_type": "task",
                "owner": "dev",
                "created_at": "2026-03-09T00:00:00Z",
                "created_by": "dev",
                "updated_at": "2026-03-09T00:00:00Z",
            }
        ],
    )

    calls = iter([tmp_path / "project", RuntimeError("second lookup fail")])

    def _load_project_directory(_root: Path):
        value = next(calls)
        if isinstance(value, Exception):
            raise value
        return value

    monkeypatch.setattr(beads_write, "load_project_directory", _load_project_directory)
    monkeypatch.setattr(
        beads_write, "load_beads_issue", lambda _root, _id: build_issue("a")
    )
    monkeypatch.setattr(beads_write, "publish_issue_mutation", lambda *_a: None)
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt-1")
    )
    monkeypatch.setattr(beads_write, "comment_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "write_events_batch", lambda *_a: None)
    beads_write.add_beads_comment(tmp_path, "a", "u", "t")


def test_descendants_and_delete_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(
        issues_path,
        [
            _record("root"),
            _record("c1", parent="root"),
            _record("c2", parent="root"),
            _record("g1", parent="c1"),
            {"parent": "root"},
        ],
    )

    assert beads_write.get_beads_descendant_identifiers(tmp_path, "root") == [
        "g1",
        "c1",
        "c2",
    ]

    missing_root = tmp_path / "missing"
    assert beads_write.get_beads_descendant_identifiers(missing_root, "x") == []

    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(beads_write, "delete_events_for_issues", lambda *_a: None)
    monkeypatch.setattr(beads_write, "publish_issue_deleted", lambda *_a: None)

    beads_write._delete_single_beads_issue(tmp_path, issues_path, "c1")
    remaining = beads_write._load_beads_records(issues_path)
    assert all(record.get("id") != "c1" for record in remaining)

    with pytest.raises(beads_write.BeadsDeleteError, match="not found"):
        beads_write._delete_single_beads_issue(tmp_path, issues_path, "missing")

    with pytest.raises(beads_write.BeadsDeleteError, match="no .beads directory"):
        beads_write.delete_beads_issue(tmp_path / "x", "a")

    beads = tmp_path / "x2" / ".beads"
    beads.mkdir(parents=True)
    with pytest.raises(beads_write.BeadsDeleteError, match="no issues.jsonl"):
        beads_write.delete_beads_issue(tmp_path / "x2", "a")


def test_delete_single_beads_issue_dependency_cleanup_and_project_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(
        issues_path,
        [
            _record("a"),
            _record(
                "b",
                parent="a",
                dependencies=[{"type": "parent-child", "depends_on_id": "a"}],
            ),
        ],
    )
    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("proj fail")),
    )
    original = issues_path.read_text(encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="proj fail"):
        beads_write._delete_single_beads_issue(tmp_path, issues_path, "a")
    assert issues_path.read_text(encoding="utf-8") == original


def test_delete_beads_issue_recursive_and_load_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(issues_path, [_record("root"), _record("child", parent="root")])

    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda _root, identifier: build_issue(identifier),
    )
    called: list[str] = []
    monkeypatch.setattr(
        beads_write, "_delete_single_beads_issue", lambda _r, _p, i: called.append(i)
    )

    beads_write.delete_beads_issue(tmp_path, "root", recursive=True)
    assert called == ["child", "root"]

    monkeypatch.setattr(
        beads_write,
        "load_beads_issue",
        lambda _root, _id: (_ for _ in ()).throw(RuntimeError("missing")),
    )
    with pytest.raises(beads_write.BeadsDeleteError, match="missing"):
        beads_write.delete_beads_issue(tmp_path, "root")


def test_add_and_remove_beads_dependency_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    _write_jsonl(
        issues_path,
        [
            _record("a", dependencies=[]),
            _record("b", parent="a", dependencies=[]),
            _record("c", parent=None, dependencies=[]),
        ],
    )

    with pytest.raises(beads_write.BeadsWriteError, match="target issue z not found"):
        beads_write.add_beads_dependency(tmp_path, "a", "z", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="own child"):
        beads_write.add_beads_dependency(tmp_path, "a", "b", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="own parent"):
        beads_write.add_beads_dependency(tmp_path, "b", "a", "blocked-by")

    beads_write.add_beads_dependency(tmp_path, "a", "c", "blocked-by")
    with pytest.raises(beads_write.BeadsWriteError, match="already exists"):
        beads_write.add_beads_dependency(tmp_path, "a", "c", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.add_beads_dependency(tmp_path, "missing", "c", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="no .beads directory"):
        beads_write.add_beads_dependency(tmp_path / "x", "a", "c", "blocked-by")
    beads = tmp_path / "x2" / ".beads"
    beads.mkdir(parents=True)
    with pytest.raises(beads_write.BeadsWriteError, match="no issues.jsonl"):
        beads_write.add_beads_dependency(tmp_path / "x2", "a", "c", "blocked-by")

    # dependencies non-list branch, and no-project return path.
    _write_jsonl(
        issues_path, [_record("a", dependencies={}), _record("c", dependencies=[])]
    )
    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )
    beads_write.add_beads_dependency(tmp_path, "a", "c", "blocked-by")

    monkeypatch.setattr(
        beads_write, "load_project_directory", lambda _root: tmp_path / "project"
    )
    monkeypatch.setattr(
        beads_write, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(beads_write, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        beads_write, "create_event", lambda **kwargs: SimpleNamespace(event_id="evt")
    )
    monkeypatch.setattr(beads_write, "dependency_payload", lambda *_a: {"ok": True})
    monkeypatch.setattr(
        beads_write, "events_dir_for_project", lambda _p: tmp_path / "events"
    )
    monkeypatch.setattr(
        beads_write,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event fail")),
    )

    original = issues_path.read_text(encoding="utf-8")
    with pytest.raises(beads_write.BeadsWriteError, match="event fail"):
        beads_write.add_beads_dependency(tmp_path, "a", "c", "relates-to")
    assert issues_path.read_text(encoding="utf-8") == original

    # remove dependency paths
    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.remove_beads_dependency(tmp_path, "missing", "c", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.remove_beads_dependency(tmp_path, "a", "c", "missing-type")

    with pytest.raises(beads_write.BeadsWriteError, match="event fail"):
        beads_write.remove_beads_dependency(tmp_path, "a", "c", "blocked-by")

    with pytest.raises(beads_write.BeadsWriteError, match="no .beads directory"):
        beads_write.remove_beads_dependency(tmp_path / "y", "a", "c", "blocked-by")
    beads_y = tmp_path / "y2" / ".beads"
    beads_y.mkdir(parents=True)
    with pytest.raises(beads_write.BeadsWriteError, match="no issues.jsonl"):
        beads_write.remove_beads_dependency(tmp_path / "y2", "a", "c", "blocked-by")

    _write_jsonl(
        issues_path, [_record("a", dependencies={}), _record("c", dependencies=[])]
    )
    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )
    with pytest.raises(beads_write.BeadsWriteError, match="not found"):
        beads_write.remove_beads_dependency(tmp_path, "a", "c", "blocked-by")

    _write_jsonl(
        issues_path,
        [
            _record(
                "a",
                dependencies=[{"type": "blocked-by", "depends_on_id": "c"}],
            ),
            _record("c", dependencies=[]),
        ],
    )
    monkeypatch.setattr(
        beads_write,
        "load_project_directory",
        lambda _root: (_ for _ in ()).throw(RuntimeError("no project")),
    )
    beads_write.remove_beads_dependency(tmp_path, "a", "c", "blocked-by")


def test_load_records_and_append_record(tmp_path: Path) -> None:
    issues_path = tmp_path / ".beads" / "issues.jsonl"
    issues_path.parent.mkdir(parents=True, exist_ok=True)
    issues_path.write_text("\n" + json.dumps(_record("a")) + "\n", encoding="utf-8")

    beads_write._append_beads_record(issues_path, _record("b"))
    records = beads_write._load_beads_records(issues_path)
    assert [record.get("id") for record in records] == ["a", "b"]



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import file_io
from kanbus.config_loader import ConfigurationError
from kanbus.project import ProjectMarkerError

from test_helpers import build_project_configuration


def test_ensure_git_repository_success_and_failure(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(
        file_io.subprocess,
        "run",
        lambda *_a, **_k: SimpleNamespace(returncode=0, stdout="true\n"),
    )
    file_io.ensure_git_repository(tmp_path)

    monkeypatch.setattr(
        file_io.subprocess,
        "run",
        lambda *_a, **_k: SimpleNamespace(returncode=1, stdout="false\n"),
    )
    with pytest.raises(file_io.InitializationError, match="not a git repository"):
        file_io.ensure_git_repository(tmp_path)


def test_initialize_project_creates_structure_and_files(tmp_path: Path) -> None:
    file_io.initialize_project(tmp_path, create_local=False)

    assert (tmp_path / "project" / "issues").is_dir()
    assert (tmp_path / "project" / "events").is_dir()
    assert (tmp_path / ".kanbus.yml").exists()
    assert (tmp_path / ".cursorignore").exists()
    assert (tmp_path / ".claude" / "settings.json").exists()
    assert (tmp_path / ".vscode" / "settings.json").exists()
    assert (tmp_path / ".gitignore").read_text(
        encoding="utf-8"
    ).strip() == "project/.overlay/"
    assert (tmp_path / file_io.DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME).exists()
    assert (
        tmp_path / "project" / "wiki" / file_io.DEFAULT_WIKI_INDEX_FILENAME
    ).exists()
    assert (
        tmp_path / "project" / "wiki" / file_io.DEFAULT_WIKI_WHATS_NEXT_FILENAME
    ).exists()


def test_initialize_project_with_existing_project_or_local(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    (tmp_path / "project").mkdir()
    with pytest.raises(file_io.InitializationError, match="already initialized"):
        file_io.initialize_project(tmp_path)

    tmp2 = tmp_path / "two"
    tmp2.mkdir()
    called: list[Path] = []
    monkeypatch.setattr(
        file_io,
        "ensure_project_local_directory",
        lambda project_dir: called.append(project_dir),
    )
    file_io.initialize_project(tmp2, create_local=True)
    assert called == [tmp2 / "project"]


def test_ensure_gitignore_entry_appends_and_dedupes(tmp_path: Path) -> None:
    file_io._ensure_gitignore_entry(tmp_path, "project/.overlay/")
    file_io._ensure_gitignore_entry(tmp_path, "project/.overlay/")
    assert (tmp_path / ".gitignore").read_text(
        encoding="utf-8"
    ) == "project/.overlay/\n"


def test_guard_file_writers(tmp_path: Path) -> None:
    subdir = tmp_path / "issues"
    subdir.mkdir()
    file_io._write_guard_files_in_subdir(subdir, "issues")
    assert (subdir / "AGENTS.md").exists()
    assert (subdir / "DO_NOT_EDIT").exists()

    project_dir = tmp_path / "project"
    (project_dir / "issues").mkdir(parents=True)
    (project_dir / "events").mkdir(parents=True)
    file_io._write_project_guard_files(project_dir)
    assert (project_dir / "AGENTS.md").exists()


def test_write_project_guard_files_if_missing_paths(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    (project_dir / "issues").mkdir(parents=True)
    (project_dir / "events").mkdir(parents=True)

    file_io._write_project_guard_files_if_missing(project_dir)
    assert (project_dir / "issues" / "AGENTS.md").exists()
    assert (project_dir / "events" / "DO_NOT_EDIT").exists()
    assert (project_dir / "AGENTS.md").exists()

    partial_dir = tmp_path / "partial"
    (partial_dir / "issues").mkdir(parents=True)
    file_io._write_project_guard_files_if_missing(partial_dir)
    assert (partial_dir / "issues" / "AGENTS.md").exists()


def test_write_tool_block_files_create_and_preserve_existing(tmp_path: Path) -> None:
    file_io._write_tool_block_files(tmp_path)

    claude_settings = tmp_path / ".claude" / "settings.json"
    vscode_settings = tmp_path / ".vscode" / "settings.json"
    assert json.loads(claude_settings.read_text(encoding="utf-8"))["permissions"][
        "deny"
    ]
    assert json.loads(vscode_settings.read_text(encoding="utf-8"))["files.exclude"]

    (tmp_path / ".cursorignore").write_text("custom\n", encoding="utf-8")
    claude_settings.write_text("custom", encoding="utf-8")
    vscode_settings.write_text("custom", encoding="utf-8")
    file_io._write_tool_block_files(tmp_path)
    assert (tmp_path / ".cursorignore").read_text(encoding="utf-8") == "custom\n"
    assert claude_settings.read_text(encoding="utf-8") == "custom"
    assert vscode_settings.read_text(encoding="utf-8") == "custom"


def test_detect_repairable_project_issues_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        file_io,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing")),
    )
    assert (
        file_io.detect_repairable_project_issues(tmp_path, allow_uninitialized=True)
        is None
    )
    with pytest.raises(ProjectMarkerError, match="missing"):
        file_io.detect_repairable_project_issues(tmp_path, allow_uninitialized=False)

    monkeypatch.setattr(
        file_io,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ConfigurationError("bad cfg")),
    )
    with pytest.raises(ConfigurationError, match="bad cfg"):
        file_io.detect_repairable_project_issues(tmp_path, allow_uninitialized=False)

    config = build_project_configuration(project_directory="project")
    config_path = tmp_path / ".kanbus.yml"
    monkeypatch.setattr(file_io, "get_configuration_path", lambda _root: config_path)
    monkeypatch.setattr(file_io, "load_project_configuration", lambda _path: config)

    missing_project = file_io.detect_repairable_project_issues(
        tmp_path, allow_uninitialized=False
    )
    assert missing_project is not None
    assert missing_project.missing_project_dir is True

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    missing_subdirs = file_io.detect_repairable_project_issues(
        tmp_path, allow_uninitialized=False
    )
    assert missing_subdirs is not None
    assert missing_subdirs.missing_issues_dir is True
    assert missing_subdirs.missing_events_dir is True

    (project_dir / "issues").mkdir()
    (project_dir / "events").mkdir()
    assert (
        file_io.detect_repairable_project_issues(tmp_path, allow_uninitialized=False)
        is None
    )


def test_repair_project_structure_and_existing_vscode(tmp_path: Path) -> None:
    plan = file_io.RepairPlan(
        project_dir=tmp_path / "project",
        missing_project_dir=True,
        missing_issues_dir=True,
        missing_events_dir=True,
    )
    file_io.repair_project_structure(tmp_path, plan)

    assert (tmp_path / "project" / "issues").is_dir()
    assert (tmp_path / "project" / "events").is_dir()
    assert (tmp_path / "project" / "AGENTS.md").exists()
    assert (tmp_path / ".vscode" / "settings.json").exists()

    existing = tmp_path / ".vscode" / "settings.json"
    existing.write_text("custom", encoding="utf-8")
    file_io.repair_project_structure(
        tmp_path,
        file_io.RepairPlan(
            project_dir=tmp_path / "project",
            missing_project_dir=False,
            missing_issues_dir=False,
            missing_events_dir=False,
        ),
    )
    assert existing.read_text(encoding="utf-8") == "custom"



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import console_snapshot
from kanbus.config_loader import ConfigurationError
from kanbus.models import OverlayConfig
from kanbus.project import ProjectMarkerError

from test_helpers import build_issue, build_project_configuration


def write_issue_file(project_dir: Path, issue_id: str, title: str) -> None:
    issue = build_issue(issue_id, title=title)
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True, exist_ok=True)
    payload = issue.model_dump(by_alias=True, mode="json")
    (issues_dir / f"{issue_id}.json").write_text(json.dumps(payload), encoding="utf-8")


def test_tag_issue_adds_project_and_source() -> None:
    issue = build_issue("kanbus-001")

    tagged = console_snapshot._tag_issue(issue, project_label="alpha", source="shared")

    assert tagged.custom["project_label"] == "alpha"
    assert tagged.custom["source"] == "shared"


def test_read_issues_from_dir_sorts_and_ignores_non_json(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    write_issue_file(project_dir, "kanbus-b", "Issue B")
    write_issue_file(project_dir, "kanbus-a", "Issue A")
    (project_dir / "issues" / "notes.txt").write_text("skip", encoding="utf-8")

    issues = console_snapshot._read_issues_from_dir(project_dir / "issues")

    assert [issue.identifier for issue in issues] == ["kanbus-a", "kanbus-b"]


def test_load_console_issues_includes_shared_and_local(tmp_path: Path) -> None:
    root = tmp_path
    project_dir = root / "project"
    local_dir = root / "project-local"
    write_issue_file(project_dir, "kanbus-shared", "Shared issue")
    write_issue_file(local_dir, "kanbus-local", "Local issue")
    configuration = build_project_configuration()

    issues = console_snapshot._load_console_issues(root, project_dir, configuration)

    by_id = {issue.identifier: issue for issue in issues}
    assert by_id["kanbus-shared"].custom["source"] == "shared"
    assert by_id["kanbus-local"].custom["source"] == "local"


def test_load_console_issues_raises_when_project_issues_missing(tmp_path: Path) -> None:
    configuration = build_project_configuration()

    try:
        console_snapshot._load_console_issues(
            tmp_path, tmp_path / "missing", configuration
        )
    except console_snapshot.ConsoleSnapshotError as error:
        assert "project/issues directory not found" in str(error)
    else:
        raise AssertionError("expected ConsoleSnapshotError")


def test_format_timestamp_uses_utc_z_suffix() -> None:
    from datetime import datetime, timezone

    value = datetime(2026, 3, 6, 12, 0, 0, 123456, tzinfo=timezone.utc)

    assert console_snapshot._format_timestamp(value) == "2026-03-06T12:00:00.123Z"


def test_build_console_snapshot_includes_config_issues_and_timestamp(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issue = build_issue("kanbus-1")
    config = build_project_configuration()
    monkeypatch.setattr(
        console_snapshot,
        "_load_project_context",
        lambda _root: (tmp_path / "project", config),
    )
    monkeypatch.setattr(
        console_snapshot,
        "_load_console_issues",
        lambda _root, _project, _config: [issue],
    )

    snapshot = console_snapshot.build_console_snapshot(tmp_path)

    assert snapshot["config"]["project_key"] == "kanbus"
    assert snapshot["issues"][0]["id"] == "kanbus-1"
    assert snapshot["updated_at"].endswith("Z")


def test_load_project_context_wraps_configuration_lookup_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        console_snapshot,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing config")),
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="missing config"):
        console_snapshot._load_project_context(tmp_path)


def test_load_project_context_wraps_configuration_validation_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config_path = tmp_path / ".kanbus.yml"
    monkeypatch.setattr(
        console_snapshot, "get_configuration_path", lambda _root: config_path
    )
    monkeypatch.setattr(
        console_snapshot,
        "load_project_configuration",
        lambda _path: (_ for _ in ()).throw(ConfigurationError("bad config")),
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="bad config"):
        console_snapshot._load_project_context(tmp_path)


def test_load_issues_with_virtual_projects_uses_beads_fallback(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    repo_root = tmp_path / "repo"
    project_dir = repo_root / "project"
    project_dir.mkdir(parents=True, exist_ok=True)
    beads_dir = repo_root / ".beads"
    beads_dir.mkdir(parents=True, exist_ok=True)
    (beads_dir / "issues.jsonl").write_text("", encoding="utf-8")
    issue = build_issue("kanbus-bdx")
    labeled = [SimpleNamespace(label="alpha", project_dir=project_dir)]
    config = build_project_configuration(
        project_key="kanbus",
        virtual_projects={"alpha": {"path": "project"}},
    )
    config = config.model_copy(
        update={"overlay": OverlayConfig(enabled=True, ttl_s=86400)}
    )

    monkeypatch.setattr(
        console_snapshot, "resolve_labeled_projects", lambda _root: labeled
    )
    monkeypatch.setattr(console_snapshot, "load_beads_issues", lambda _repo: [issue])

    issues = console_snapshot._load_issues_with_virtual_projects(repo_root, config)

    assert len(issues) == 1
    assert issues[0].identifier == "kanbus-bdx"
    assert issues[0].custom["project_label"] == "alpha"
    assert issues[0].custom["source"] == "shared"


def test_get_issues_for_root_delegates_to_console_loader(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issue = build_issue("kanbus-42")
    config = build_project_configuration()
    monkeypatch.setattr(
        console_snapshot,
        "_load_project_context",
        lambda _root: (tmp_path / "project", config),
    )
    monkeypatch.setattr(
        console_snapshot,
        "_load_console_issues",
        lambda _root, _project_dir, _cfg: [issue],
    )

    issues = console_snapshot.get_issues_for_root(tmp_path)

    assert len(issues) == 1
    assert issues[0].identifier == "kanbus-42"


def test_load_console_issues_uses_beads_mode_and_sorts(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    configuration = build_project_configuration(beads_compatibility=True)
    issue_b = build_issue("kanbus-b")
    issue_a = build_issue("kanbus-a")
    monkeypatch.setattr(
        console_snapshot, "load_beads_issues", lambda _root: [issue_b, issue_a]
    )

    issues = console_snapshot._load_console_issues(tmp_path, project_dir, configuration)

    assert [issue.identifier for issue in issues] == ["kanbus-a", "kanbus-b"]


def test_load_console_issues_wraps_beads_migration_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    configuration = build_project_configuration(beads_compatibility=True)
    monkeypatch.setattr(
        console_snapshot,
        "load_beads_issues",
        lambda _root: (_ for _ in ()).throw(
            console_snapshot.MigrationError("beads broken")
        ),
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="beads broken"):
        console_snapshot._load_console_issues(tmp_path, project_dir, configuration)


def test_load_console_issues_wraps_invalid_issue_file_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    (project_dir / "issues").mkdir(parents=True, exist_ok=True)
    configuration = build_project_configuration()
    monkeypatch.setattr(
        console_snapshot,
        "_read_issues_from_dir",
        lambda _path: (_ for _ in ()).throw(ValueError("bad issue file")),
    )

    with pytest.raises(
        console_snapshot.ConsoleSnapshotError, match="issue file is invalid"
    ):
        console_snapshot._load_console_issues(tmp_path, project_dir, configuration)


def test_load_issues_with_virtual_projects_wraps_resolve_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config = build_project_configuration(
        virtual_projects={"alpha": {"path": "project"}}
    )
    monkeypatch.setattr(
        console_snapshot,
        "resolve_labeled_projects",
        lambda _root: (_ for _ in ()).throw(RuntimeError("resolve failed")),
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="resolve failed"):
        console_snapshot._load_issues_with_virtual_projects(tmp_path, config)


def test_load_issues_with_virtual_projects_wraps_beads_migration_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    repo_root = tmp_path / "repo"
    project_dir = repo_root / "project"
    project_dir.mkdir(parents=True, exist_ok=True)
    beads_dir = repo_root / ".beads"
    beads_dir.mkdir(parents=True, exist_ok=True)
    (beads_dir / "issues.jsonl").write_text("", encoding="utf-8")
    labeled = [SimpleNamespace(label="alpha", project_dir=project_dir)]
    config = build_project_configuration(
        virtual_projects={"alpha": {"path": "project"}}
    )

    monkeypatch.setattr(
        console_snapshot, "resolve_labeled_projects", lambda _root: labeled
    )
    monkeypatch.setattr(
        console_snapshot,
        "load_beads_issues",
        lambda _repo: (_ for _ in ()).throw(
            console_snapshot.MigrationError("bad beads")
        ),
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="bad beads"):
        console_snapshot._load_issues_with_virtual_projects(repo_root, config)


def test_load_project_context_success_and_config_path_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    config_path = tmp_path / ".kanbus.yml"
    config = build_project_configuration(project_directory="project")
    monkeypatch.setattr(
        console_snapshot, "get_configuration_path", lambda _root: config_path
    )
    monkeypatch.setattr(
        console_snapshot, "load_project_configuration", lambda _p: config
    )
    project_dir, loaded = console_snapshot._load_project_context(tmp_path)
    assert project_dir == tmp_path / "project"
    assert loaded.project_key == config.project_key

    monkeypatch.setattr(
        console_snapshot,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ConfigurationError("bad path")),
    )
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="bad path"):
        console_snapshot._load_project_context(tmp_path)


def test_load_console_issues_virtual_and_not_dir_and_permission_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    project_dir.mkdir(parents=True)
    cfg_virtual = build_project_configuration(
        virtual_projects={"alpha": {"path": "project"}}
    )
    issue = build_issue("kanbus-v")
    monkeypatch.setattr(
        console_snapshot, "_load_issues_with_virtual_projects", lambda *_a: [issue]
    )
    assert console_snapshot._load_console_issues(
        tmp_path, project_dir, cfg_virtual
    ) == [issue]

    issues_path = project_dir / "issues"
    issues_path.write_text("not a dir", encoding="utf-8")
    cfg = build_project_configuration()
    with pytest.raises(
        console_snapshot.ConsoleSnapshotError,
        match="project/issues directory not found",
    ):
        console_snapshot._load_console_issues(tmp_path, project_dir, cfg)

    issues_path.unlink()
    issues_path.mkdir()
    monkeypatch.setattr(
        console_snapshot,
        "_read_issues_from_dir",
        lambda _p: (_ for _ in ()).throw(PermissionError("no read")),
    )
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="no read"):
        console_snapshot._load_console_issues(tmp_path, project_dir, cfg)


def test_load_console_issues_local_read_permission_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    local_dir = tmp_path / "project-local"
    (project_dir / "issues").mkdir(parents=True)
    (local_dir / "issues").mkdir(parents=True)
    cfg = build_project_configuration()
    shared_issue = build_issue("kanbus-shared")

    def fake_read(path: Path):
        if "project-local" in str(path):
            raise PermissionError("local denied")
        return [shared_issue]

    monkeypatch.setattr(console_snapshot, "_read_issues_from_dir", fake_read)
    monkeypatch.setattr(
        console_snapshot, "find_project_local_directory", lambda _p: local_dir
    )

    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="local denied"):
        console_snapshot._load_console_issues(tmp_path, project_dir, cfg)


def test_load_console_issues_local_read_invalid_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    local_dir = tmp_path / "project-local"
    (project_dir / "issues").mkdir(parents=True)
    (local_dir / "issues").mkdir(parents=True)
    cfg = build_project_configuration()
    shared_issue = build_issue("kanbus-shared")

    def fake_read(path: Path):
        if "project-local" in str(path):
            raise ValueError("bad local")
        return [shared_issue]

    monkeypatch.setattr(console_snapshot, "_read_issues_from_dir", fake_read)
    monkeypatch.setattr(
        console_snapshot, "find_project_local_directory", lambda _p: local_dir
    )

    with pytest.raises(
        console_snapshot.ConsoleSnapshotError, match="issue file is invalid"
    ):
        console_snapshot._load_console_issues(tmp_path, project_dir, cfg)


def test_load_issues_with_virtual_projects_shared_local_and_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    repo = tmp_path / "repo"
    p1 = repo / "alpha" / "project"
    p2 = repo / "beta" / "project"
    (p1 / "issues").mkdir(parents=True)
    (p1.parent / "project-local" / "issues").mkdir(parents=True)
    p2.mkdir(parents=True)
    (p2.parent / ".beads").mkdir(parents=True)
    (p2.parent / ".beads" / "issues.jsonl").write_text("", encoding="utf-8")

    labeled = [
        SimpleNamespace(label="alpha", project_dir=p1),
        SimpleNamespace(label="beta", project_dir=p2),
    ]
    cfg = build_project_configuration(
        virtual_projects={"alpha": {"path": str(p1)}, "beta": {"path": str(p2)}}
    )
    shared_issue = build_issue("kanbus-a1")
    local_issue = build_issue("kanbus-a2")
    beads_issue = build_issue("kanbus-b1")

    def fake_read(path: Path):
        if path == p1 / "issues":
            return [shared_issue]
        if path == p1.parent / "project-local" / "issues":
            return [local_issue]
        return []

    monkeypatch.setattr(
        console_snapshot, "resolve_labeled_projects", lambda _r: labeled
    )
    monkeypatch.setattr(console_snapshot, "_read_issues_from_dir", fake_read)
    monkeypatch.setattr(
        console_snapshot,
        "find_project_local_directory",
        lambda project: p1.parent / "project-local" if project == p1 else None,
    )
    monkeypatch.setattr(
        console_snapshot, "load_beads_issues", lambda _repo: [beads_issue]
    )

    issues = console_snapshot._load_issues_with_virtual_projects(repo, cfg)
    assert [issue.identifier for issue in issues] == [
        "kanbus-a1",
        "kanbus-a2",
        "kanbus-b1",
    ]
    assert {issue.custom.get("project_label") for issue in issues} == {"alpha", "beta"}

    monkeypatch.setattr(
        console_snapshot,
        "_read_issues_from_dir",
        lambda _p: (_ for _ in ()).throw(PermissionError("shared denied")),
    )
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="shared denied"):
        console_snapshot._load_issues_with_virtual_projects(repo, cfg)

    monkeypatch.setattr(
        console_snapshot,
        "_read_issues_from_dir",
        lambda _p: (_ for _ in ()).throw(ValueError("shared invalid")),
    )
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="shared invalid"):
        console_snapshot._load_issues_with_virtual_projects(repo, cfg)


def test_load_issues_with_virtual_projects_local_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    repo = tmp_path / "repo"
    project_dir = repo / "alpha" / "project"
    (project_dir / "issues").mkdir(parents=True)
    local_issues_dir = project_dir.parent / "project-local" / "issues"
    local_issues_dir.mkdir(parents=True)

    labeled = [SimpleNamespace(label="alpha", project_dir=project_dir)]
    cfg = build_project_configuration(
        virtual_projects={"alpha": {"path": str(project_dir)}}
    )
    shared_issue = build_issue("kanbus-a1")

    def permission_local(path: Path):
        if path == local_issues_dir:
            raise PermissionError("local denied")
        return [shared_issue]

    monkeypatch.setattr(
        console_snapshot, "resolve_labeled_projects", lambda _r: labeled
    )
    monkeypatch.setattr(
        console_snapshot,
        "find_project_local_directory",
        lambda _p: project_dir.parent / "project-local",
    )
    monkeypatch.setattr(console_snapshot, "_read_issues_from_dir", permission_local)
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="local denied"):
        console_snapshot._load_issues_with_virtual_projects(repo, cfg)

    def invalid_local(path: Path):
        if path == local_issues_dir:
            raise ValueError("local bad")
        return [shared_issue]

    monkeypatch.setattr(console_snapshot, "_read_issues_from_dir", invalid_local)
    with pytest.raises(console_snapshot.ConsoleSnapshotError, match="local bad"):
        console_snapshot._load_issues_with_virtual_projects(repo, cfg)



from __future__ import annotations

from pathlib import Path

import pytest

from kanbus import ids, project_management_template, text_editor, wiki_templates


def test_id_uuid_sequence_and_generation() -> None:
    ids.set_test_uuid_sequence(["abc123", "def456"])
    assert ids._next_uuid_value() == "abc123"
    assert ids._next_uuid_value() == "def456"
    ids.set_test_uuid_sequence(None)

    req = ids.IssueIdentifierRequest(title="x", prefix="kanbus", existing_ids=set())
    result = ids.generate_issue_identifier(req)
    assert result.identifier.startswith("kanbus-")


def test_format_issue_key_variants() -> None:
    assert ids.format_issue_key("12345", project_context=False) == "12345"
    assert (
        ids.format_issue_key("kanbus-abcdef123456", project_context=False)
        == "kanbus-abcdef"
    )
    assert ids.format_issue_key("kanbus-abcdef123456", project_context=True) == "abcdef"
    assert (
        ids.format_issue_key("kanbus-abcdef123456.2", project_context=False)
        == "kanbus-abcdef.2"
    )
    assert ids.format_issue_key("abcdef123456", project_context=False) == "abcdef"


def test_matches_issue_identifier_paths() -> None:
    full = "kanbus-abcdef123456"
    assert ids.matches_issue_identifier(full, full) is True
    assert (
        ids.matches_issue_identifier(
            ids.format_issue_key(full, project_context=False), full
        )
        is True
    )
    assert (
        ids.matches_issue_identifier(
            ids.format_issue_key(full, project_context=True), full
        )
        is True
    )
    assert ids.matches_issue_identifier("x" * len(full), full) is False
    assert ids.matches_issue_identifier("kanbus-ab", full) is True
    assert ids.matches_issue_identifier("zzz", full) is False


def test_generate_issue_identifier_collision_failure() -> None:
    ids.set_test_uuid_sequence(["dup"] * 10)
    req = ids.IssueIdentifierRequest(
        title="x", prefix="kanbus", existing_ids={"kanbus-dup"}
    )
    with pytest.raises(RuntimeError, match="unable to generate unique id"):
        ids.generate_issue_identifier(req)
    ids.set_test_uuid_sequence(None)


def test_generate_many_identifiers() -> None:
    ids.set_test_uuid_sequence(["a", "b", "c"])
    generated = ids.generate_many_identifiers("title", "kanbus", 3)
    assert generated == {"kanbus-a", "kanbus-b", "kanbus-c"}
    ids.set_test_uuid_sequence(None)


def test_text_editor_view_and_directory(tmp_path: Path) -> None:
    root = tmp_path
    subdir = root / "d"
    subdir.mkdir()
    (subdir / "z.txt").write_text("z", encoding="utf-8")
    (subdir / "a").mkdir()

    listing = text_editor.edit_view(root, Path("d"))
    assert listing.splitlines() == ["a/", "z.txt"]

    file_path = root / "file.txt"
    file_path.write_text("one\ntwo\nthree\n", encoding="utf-8")
    viewed = text_editor.edit_view(root, Path("file.txt"))
    assert "1: one" in viewed
    ranged = text_editor.edit_view(root, Path("file.txt"), view_range=(2, -1))
    assert ranged == "2: two\n3: three"
    from_zero = text_editor.edit_view(root, Path("file.txt"), view_range=(0, 1))
    assert from_zero == "1: one"


def test_text_editor_view_errors(tmp_path: Path) -> None:
    root = tmp_path
    with pytest.raises(text_editor.TextEditorError, match="escapes repository root"):
        text_editor.edit_view(root, Path("../x"))
    with pytest.raises(
        text_editor.TextEditorError, match="file or directory not found"
    ):
        text_editor.edit_view(root, Path("missing.txt"))


def test_text_editor_str_replace_paths(tmp_path: Path) -> None:
    root = tmp_path
    file_path = root / "replace.txt"
    file_path.write_text("hello world", encoding="utf-8")

    assert "Successfully replaced" in text_editor.edit_str_replace(
        root, Path("replace.txt"), "hello", "hi"
    )
    assert file_path.read_text(encoding="utf-8") == "hi world"

    with pytest.raises(text_editor.TextEditorError, match="escapes repository root"):
        text_editor.edit_str_replace(root, Path("../x"), "a", "b")
    with pytest.raises(text_editor.TextEditorError, match="file not found"):
        text_editor.edit_str_replace(root, Path("missing.txt"), "a", "b")

    file_path.write_text("x", encoding="utf-8")
    with pytest.raises(text_editor.TextEditorError, match="no match found"):
        text_editor.edit_str_replace(root, Path("replace.txt"), "a", "b")

    file_path.write_text("x\nx", encoding="utf-8")
    with pytest.raises(text_editor.TextEditorError, match="found 2 matches"):
        text_editor.edit_str_replace(root, Path("replace.txt"), "x", "y")


def test_text_editor_create_and_insert(tmp_path: Path) -> None:
    root = tmp_path
    created = text_editor.edit_create(root, Path("new/file.txt"), "hello")
    assert "Successfully created file" in created
    assert (root / "new" / "file.txt").read_text(encoding="utf-8") == "hello"

    with pytest.raises(text_editor.TextEditorError, match="escapes repository root"):
        text_editor.edit_create(root, Path("../x"), "x")
    with pytest.raises(text_editor.TextEditorError, match="file already exists"):
        text_editor.edit_create(root, Path("new/file.txt"), "x")

    target = root / "insert.txt"
    target.write_text("a\nb\n", encoding="utf-8")
    assert "Successfully inserted" in text_editor.edit_insert(
        root, Path("insert.txt"), 1, "x"
    )
    assert target.read_text(encoding="utf-8") == "a\nx\nb\n"

    empty = root / "empty.txt"
    empty.write_text("", encoding="utf-8")
    text_editor.edit_insert(root, Path("empty.txt"), 0, "")
    assert empty.read_text(encoding="utf-8") == ""

    with pytest.raises(text_editor.TextEditorError, match="escapes repository root"):
        text_editor.edit_insert(root, Path("../x"), 0, "x")
    with pytest.raises(text_editor.TextEditorError, match="file not found"):
        text_editor.edit_insert(root, Path("missing.txt"), 0, "x")
    with pytest.raises(text_editor.TextEditorError, match="non-negative"):
        text_editor.edit_insert(root, Path("insert.txt"), -1, "x")
    with pytest.raises(text_editor.TextEditorError, match="exceeds file length"):
        text_editor.edit_insert(root, Path("insert.txt"), 999, "x")


def test_template_constants_are_expected() -> None:
    assert wiki_templates.DEFAULT_WIKI_INDEX_FILENAME == "index.md"
    assert wiki_templates.DEFAULT_WIKI_WHATS_NEXT_FILENAME == "whats-next.md"
    assert "What's Next" in wiki_templates.DEFAULT_WIKI_INDEX
    assert 'query(status="open"' in wiki_templates.DEFAULT_WIKI_WHATS_NEXT

    assert project_management_template.DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME.endswith(
        ".template.md"
    )
    assert (
        "This is The Way."
        in project_management_template.DEFAULT_PROJECT_MANAGEMENT_TEMPLATE
    )
    assert (
        "{{ project_key }}"
        in project_management_template.DEFAULT_PROJECT_MANAGEMENT_TEMPLATE
    )



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import issue_transfer
from kanbus.issue_lookup import IssueLookupError

from test_helpers import build_issue


def write_issue(path: Path, issue_id: str) -> None:
    issue = build_issue(issue_id)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(
        json.dumps(issue.model_dump(by_alias=True, mode="json")),
        encoding="utf-8",
    )


def test_promote_issue_moves_local_to_shared_and_emits_event(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    local_dir = root / "project-local"
    local_issue = local_dir / "issues" / "kanbus-1.json"
    shared_issue = project_dir / "issues" / "kanbus-1.json"
    shared_issue.parent.mkdir(parents=True, exist_ok=True)
    write_issue(local_issue, "kanbus-1")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    monkeypatch.setattr(
        issue_transfer, "find_project_local_directory", lambda _p: local_dir
    )
    monkeypatch.setattr(issue_transfer, "get_current_user", lambda: "tester")
    calls: dict[str, object] = {}
    monkeypatch.setattr(
        issue_transfer,
        "publish_issue_mutation",
        lambda *_args: calls.setdefault("publish_mutation", True),
    )

    issue = issue_transfer.promote_issue(root, "kanbus-1")

    assert issue.identifier == "kanbus-1"
    assert shared_issue.exists()
    assert not local_issue.exists()
    assert calls.get("publish_mutation") is True
    event_files = list((project_dir / "events").glob("*.json"))
    assert len(event_files) == 1
    payload = json.loads(event_files[0].read_text(encoding="utf-8"))
    assert payload["event_type"] == "issue_promoted"


def test_promote_issue_rolls_back_when_event_write_fails(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    local_dir = root / "project-local"
    local_issue = local_dir / "issues" / "kanbus-2.json"
    shared_issue = project_dir / "issues" / "kanbus-2.json"
    shared_issue.parent.mkdir(parents=True, exist_ok=True)
    write_issue(local_issue, "kanbus-2")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    monkeypatch.setattr(
        issue_transfer, "find_project_local_directory", lambda _p: local_dir
    )
    monkeypatch.setattr(
        issue_transfer,
        "write_events_batch",
        lambda *_args, **_kwargs: (_ for _ in ()).throw(RuntimeError("boom")),
    )

    with pytest.raises(issue_transfer.IssueTransferError, match="boom"):
        issue_transfer.promote_issue(root, "kanbus-2")

    assert local_issue.exists()
    assert not shared_issue.exists()


def test_promote_issue_error_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: (_ for _ in ()).throw(IssueLookupError("missing")),
    )
    with pytest.raises(issue_transfer.IssueTransferError, match="missing"):
        issue_transfer.promote_issue(root, "kanbus-missing")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    monkeypatch.setattr(issue_transfer, "find_project_local_directory", lambda _p: None)
    with pytest.raises(issue_transfer.IssueTransferError, match="not found"):
        issue_transfer.promote_issue(root, "kanbus-missing")

    local_dir = root / "project-local"
    monkeypatch.setattr(
        issue_transfer, "find_project_local_directory", lambda _p: local_dir
    )
    with pytest.raises(issue_transfer.IssueTransferError, match="not found"):
        issue_transfer.promote_issue(root, "kanbus-missing")

    (local_dir / "issues").mkdir(parents=True, exist_ok=True)
    write_issue(local_dir / "issues" / "kanbus-dupe.json", "kanbus-dupe")
    target = project_dir / "issues" / "kanbus-dupe.json"
    target.parent.mkdir(parents=True, exist_ok=True)
    write_issue(target, "kanbus-dupe")
    with pytest.raises(issue_transfer.IssueTransferError, match="already exists"):
        issue_transfer.promote_issue(root, "kanbus-dupe")


def test_localize_issue_moves_shared_to_local_and_emits_event(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    local_dir = root / "project-local"
    shared_issue = project_dir / "issues" / "kanbus-3.json"
    local_issue = local_dir / "issues" / "kanbus-3.json"
    local_issue.parent.mkdir(parents=True, exist_ok=True)
    write_issue(shared_issue, "kanbus-3")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    monkeypatch.setattr(
        issue_transfer, "ensure_project_local_directory", lambda _p: local_dir
    )
    monkeypatch.setattr(issue_transfer, "get_current_user", lambda: "tester")
    calls: dict[str, object] = {}
    monkeypatch.setattr(
        issue_transfer,
        "publish_issue_deleted",
        lambda *_args: calls.setdefault("publish_deleted", True),
    )

    issue = issue_transfer.localize_issue(root, "kanbus-3")

    assert issue.identifier == "kanbus-3"
    assert local_issue.exists()
    assert not shared_issue.exists()
    assert calls.get("publish_deleted") is True
    event_files = list((local_dir / "events").glob("*.json"))
    assert len(event_files) == 1
    payload = json.loads(event_files[0].read_text(encoding="utf-8"))
    assert payload["event_type"] == "issue_localized"


def test_localize_issue_rolls_back_when_event_write_fails(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    local_dir = root / "project-local"
    shared_issue = project_dir / "issues" / "kanbus-4.json"
    local_issue = local_dir / "issues" / "kanbus-4.json"
    local_issue.parent.mkdir(parents=True, exist_ok=True)
    write_issue(shared_issue, "kanbus-4")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    monkeypatch.setattr(
        issue_transfer, "ensure_project_local_directory", lambda _p: local_dir
    )
    monkeypatch.setattr(
        issue_transfer,
        "write_events_batch",
        lambda *_args, **_kwargs: (_ for _ in ()).throw(RuntimeError("boom")),
    )

    with pytest.raises(issue_transfer.IssueTransferError, match="boom"):
        issue_transfer.localize_issue(root, "kanbus-4")

    assert shared_issue.exists()
    assert not local_issue.exists()


def test_localize_issue_error_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: (_ for _ in ()).throw(IssueLookupError("missing")),
    )
    with pytest.raises(issue_transfer.IssueTransferError, match="missing"):
        issue_transfer.localize_issue(root, "kanbus-missing")

    monkeypatch.setattr(
        issue_transfer,
        "load_issue_from_project",
        lambda _root, _id: SimpleNamespace(project_dir=project_dir),
    )
    with pytest.raises(
        issue_transfer.IssueTransferError, match="issue is not in shared project"
    ):
        issue_transfer.localize_issue(root, "kanbus-missing")

    shared_issue = project_dir / "issues" / "kanbus-5.json"
    local_dir = root / "project-local"
    local_issue = local_dir / "issues" / "kanbus-5.json"
    write_issue(shared_issue, "kanbus-5")
    write_issue(local_issue, "kanbus-5")
    monkeypatch.setattr(
        issue_transfer, "ensure_project_local_directory", lambda _p: local_dir
    )
    with pytest.raises(issue_transfer.IssueTransferError, match="already exists"):
        issue_transfer.localize_issue(root, "kanbus-5")



"""Console UI state: load and save the cached last-pushed UI route."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Optional

from kanbus.project import load_project_directory


def get_console_state_path(root: Path) -> Path:
    """Return the console UI state cache file path.

    :param root: Repository root path.
    :type root: Path
    :return: Path to console_state.json.
    :rtype: Path
    """
    project_dir = load_project_directory(root)
    return project_dir / ".cache" / "console_state.json"


def fetch_console_ui_state(root: Path, port: Optional[int] = None) -> Optional[dict]:  # type: ignore[type-arg]
    """Fetch the current UI state from the running console server.

    Returns ``None`` if the server is not reachable.

    :param root: Repository root path.
    :type root: Path
    :param port: HTTP port override; reads from project config if not given.
    :type port: int, optional
    :return: UI state dict or None.
    :rtype: dict or None
    """
    import urllib.error
    import urllib.request

    if port is None:
        try:
            from kanbus.project import get_configuration_path
            from kanbus.config_loader import load_project_configuration

            config_path = get_configuration_path(root)
            config = load_project_configuration(config_path)
            port = getattr(config, "console_port", None) or 5174
        except Exception:
            port = 5174

    url = f"http://127.0.0.1:{port}/api/ui-state"
    try:
        req = urllib.request.urlopen(url, timeout=3)  # noqa: S310
        return json.loads(req.read().decode())
    except (urllib.error.URLError, OSError, json.JSONDecodeError):
        return None



from __future__ import annotations

import json
import sys
from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[2]
sys.path.insert(0, str(REPO_ROOT / "tools"))

import coverage_changed_files_gate  # noqa: E402


def write_coverage_xml(path: Path, entries: dict[str, tuple[int, int]]) -> None:
    classes = []
    for filename, (covered, total) in entries.items():
        line_nodes = []
        for index in range(total):
            hits = 1 if index < covered else 0
            line_nodes.append(f'<line number="{index + 1}" hits="{hits}"/>')
        classes.append(
            f'<class filename="{filename}" line-rate="0"><lines>{"".join(line_nodes)}</lines></class>'
        )
    xml = (
        '<coverage line-rate="0"><packages><package><classes>'
        f'{"".join(classes)}'
        "</classes></package></packages></coverage>"
    )
    path.write_text(xml, encoding="utf-8")


def run_main(
    tmp_path: Path,
    monkeypatch,
    changed_files: list[str],
    python_entries: dict[str, tuple[int, int]],
    rust_entries: dict[str, tuple[int, int]],
    config: dict[str, object],
) -> int:
    python_xml = tmp_path / "python.xml"
    rust_xml = tmp_path / "rust.xml"
    config_path = tmp_path / "gate.json"
    write_coverage_xml(python_xml, python_entries)
    write_coverage_xml(rust_xml, rust_entries)
    config_path.write_text(json.dumps(config), encoding="utf-8")
    monkeypatch.setattr(
        coverage_changed_files_gate,
        "_run_git_diff",
        lambda base_ref, head_ref: changed_files,
    )
    return coverage_changed_files_gate.main(
        [
            "--python-xml",
            str(python_xml),
            "--rust-xml",
            str(rust_xml),
            "--config",
            str(config_path),
            "--base-ref",
            "BASE",
            "--head-ref",
            "HEAD",
            "--json",
        ]
    )


def test_gate_passes_when_no_changed_source_files(tmp_path: Path, monkeypatch) -> None:
    exit_code = run_main(
        tmp_path=tmp_path,
        monkeypatch=monkeypatch,
        changed_files=["README.md"],
        python_entries={"kanbus/cli.py": (80, 100)},
        rust_entries={"src/cli.rs": (80, 100)},
        config={
            "python_min": 75.0,
            "rust_min": 65.0,
            "waivers": [],
            "waiver_min_improvement_points": 3.0,
            "file_baselines": {},
        },
    )
    assert exit_code == 0


def test_gate_fails_for_non_waived_file_below_minimum(
    tmp_path: Path, monkeypatch
) -> None:
    exit_code = run_main(
        tmp_path=tmp_path,
        monkeypatch=monkeypatch,
        changed_files=["python/src/kanbus/cli.py"],
        python_entries={"kanbus/cli.py": (60, 100)},
        rust_entries={"src/cli.rs": (80, 100)},
        config={
            "python_min": 75.0,
            "rust_min": 65.0,
            "waivers": [],
            "waiver_min_improvement_points": 3.0,
            "file_baselines": {},
        },
    )
    assert exit_code == 1


def test_gate_passes_waived_file_with_required_improvement(
    tmp_path: Path, monkeypatch
) -> None:
    exit_code = run_main(
        tmp_path=tmp_path,
        monkeypatch=monkeypatch,
        changed_files=["python/src/kanbus/snyk_sync.py"],
        python_entries={"kanbus/snyk_sync.py": (40, 100)},
        rust_entries={"src/cli.rs": (80, 100)},
        config={
            "python_min": 75.0,
            "rust_min": 65.0,
            "waivers": ["python/src/kanbus/snyk_sync.py"],
            "waiver_min_improvement_points": 3.0,
            "file_baselines": {"python/src/kanbus/snyk_sync.py": 35.0},
        },
    )
    assert exit_code == 0


def test_gate_fails_waived_file_with_insufficient_improvement(
    tmp_path: Path, monkeypatch
) -> None:
    exit_code = run_main(
        tmp_path=tmp_path,
        monkeypatch=monkeypatch,
        changed_files=["python/src/kanbus/snyk_sync.py"],
        python_entries={"kanbus/snyk_sync.py": (36, 100)},
        rust_entries={"src/cli.rs": (80, 100)},
        config={
            "python_min": 75.0,
            "rust_min": 65.0,
            "waivers": ["python/src/kanbus/snyk_sync.py"],
            "waiver_min_improvement_points": 3.0,
            "file_baselines": {"python/src/kanbus/snyk_sync.py": 35.0},
        },
    )
    assert exit_code == 1
